import json
import os
import ssl

import pika

from .base import BaseMessageBrokerConnector


class RabbitMQConnector(BaseMessageBrokerConnector):
    def __init__(self, queue_declare, **kwargs):
        super().__init__(**kwargs)

        # * old version
        # credentials = pika.PlainCredentials(
        #     os.getenv("RABBIT_MQ_USER"), os.getenv("RABBIT_MQ_PASSWORD")
        # )
        # parameters = pika.ConnectionParameters(
        #     os.getenv("RABBIT_MQ_CONN"), 5672, os.getenv("RABBIT_MQ_VHOST"), credentials
        # )

        # *new version
        if os.getenv("RABBIT_MQ_CONN_TYPE") == "PARAMETER":
            credentials = pika.PlainCredentials(
                os.getenv("RABBIT_MQ_USER"), os.getenv("RABBIT_MQ_PASSWORD")
            )
            parameters = pika.ConnectionParameters(
                os.getenv("RABBIT_MQ_CONN"),
                os.getenv("RABBIT_MQ_PORT"),
                os.getenv("RABBIT_MQ_VHOST"),
                credentials,
            )
        elif os.getenv("RABBIT_MQ_CONN_TYPE") == "STRING":
            parameters = pika.URLParameters(os.getenv("RABBIT_MQ_CONN_STRING"))

        if os.getenv("RABBIT_MQ_CONN_SSL") == "ENABLE":
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            ssl_context.set_ciphers("ECDHE+AESGCM:!ECDSA")
            parameters.ssl_options = pika.SSLOptions(context=ssl_context)

        self.rabbit_mq_conn = pika.BlockingConnection(parameters)

        self.channel = self.rabbit_mq_conn.channel()
        self.channel.queue_declare(queue=queue_declare, durable=True)

    def close(self):
        if self.rabbit_mq_conn:
            self.rabbit_mq_conn.close()

    def send(self, message, target, **kwargs):
        self.channel.basic_publish(
            exchange="",
            routing_key=target,
            body=json.dumps(message),
            properties=pika.BasicProperties(
                delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE
            ),
        )

    def receive(self, receive_from, received_callback, **kwargs):
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(
            queue=receive_from, on_message_callback=received_callback, auto_ack=False
        )

        self.channel.start_consuming()

    def send_processed_message_status(self, is_processed=True, **kwargs):
        if is_processed:
            return self.channel.basic_ack(delivery_tag=kwargs["delivery_tag"])
        else:
            return self.channel.basic_nack(delivery_tag=kwargs["delivery_tag"])



class BaseDatabaseConnector:

    def __init__(self, connection_string):
        self.database = None
        self.connection = self.get_connection(connection_string=connection_string)

    def close(self):
        pass

    @staticmethod
    def get_connection(**kwargs):
        pass

    def get(self, key, obj_type, **kwargs):
        pass

    def get_list(self, criteria, obj_type, **kwargs):
        pass

    def insert(self, obj, obj_type, **kwargs):
        pass

    def update(self, criteria, new_obj, obj_type, **kwargs):
        pass



from ..items.website_item import WebsiteItem
from .base import BaseSpider


class WebsiteSpider(BaseSpider):
    instance = "website"

    def __init__(self, url):
        super().__init__(url)

    @staticmethod
    def get_new_item_instance():
        return WebsiteItem()

    def set_default_value(self):
        items = super().set_default_value()
        items["name"] = ""  # update when error
        items["css_selector"] = ""
        items["categories_num"] = []
        items["total_categories"] = 0  # update when error

        # use in pipeline to create categories and send message
        items["category_urls"] = []
        return items



import logging
import os
import sys
import time
import traceback
from datetime import datetime, timedelta
from pymongo import read_concern, write_concern

import graypy
from crontab import CronTab

from ...base_crawler import BaseCrawler
from ...connectors.databases.mongodb_connector import MongoDBConnector
from ...log.default import get_default_log


class WebsiteCrawler(BaseCrawler):
    def __init__(self):
        self.receive_queue = None
        self.transmit_queue = None

        self.website_coll = None
        self.categories_coll = None
        self.logger = None

        super().__init__()

        self.cron_cmd_template = os.getenv("CRON_CMD_TEMPLATE")
        self.user = os.getenv("USER")
        self.loop_time = int(os.getenv("LOOP_TIME"))
        self.retry_time_limit = os.getenv("RETRY_TIME_LIMIT")

    def setup_database(self, reinit=False):
        super().setup_database()
        self.website_coll = self.database[f'{os.getenv("WEBSITE_COLL")}']
        self.categories_coll = self.database[f'{os.getenv("CATEGORIES_COLL")}']

    def load_database_env(self):
        super().load_database_env()

    # def load_connector_env(self):
    #     super().load_connector_env()
    #     self.receive_queue = os.getenv("RECEIVE_QUEUE")
    #     self.transmit_queue = os.getenv("TRANSMIT_QUEUE")

    def init_database_connector(self):
        self.database_client = MongoDBConnector(
            connection_string=self.database_params["conn_string"],
            default_database=self.database_params["database"],
        )
        self.database = self.database_client.connection[
            self.database_params["database"]
        ]

    def get_logger(self):
        if not self.logger:
            self.logger = get_default_log()

        return self.logger

    def run(self):
        super().run()
        first_time = True

        next_time = datetime.now() + timedelta(seconds=self.loop_time)

        while True:
            if first_time:
                self.logger.info("[*] Website instance: Init first time")

                self.remove_all_job()

                self.logger.info("[*] Website instance: Remove all prev job")

                self.init_schedule()

                first_time = False
            else:
                date_time_now = datetime.now()

                if date_time_now >= next_time:
                    self.init_schedule()
                    next_time = date_time_now + timedelta(seconds=self.loop_time)
                else:
                    # put process to sleep for remained time to prevent continuous loop
                    remained_time = next_time - date_time_now
                    if remained_time.seconds > 0:
                        time.sleep((remained_time + timedelta(seconds=5)).seconds)

    def remove_all_job(self):
        cron = CronTab(user=self.user)
        cron.remove_all()
        cron.write()

    def init_schedule(self):
        connection = None
        rc_snapshot = read_concern.ReadConcern("snapshot")
        wc_majority = write_concern.WriteConcern("majority")
        try:
            self.init_database_connector()
            connection = self.database_client.connection

            cron = CronTab(user=self.user)

            collection = self.website_coll

            for website in collection.find():
                found = False
                for job in cron:
                    if job.comment == website["spider_name"]:
                        found = True

                        # ! check status still in-active (error when spider not run) if err_re_process > 2 time => omit
                        if website["status"] == "in-active":
                            if website["err_re_process"] > int(self.retry_time_limit):
                                self.get_logger().info(
                                    f"[*] Remove cronjob of {website['spider_name']}: exceed retry time when process job"
                                )
                                cron.remove(job)
                                cron.write()
                                return
                            # update err_re_process
                            with connection.start_session(
                                {"causalConsistency": False}
                            ) as session:
                                session.with_transaction(
                                    lambda s: self.update_website_callback(
                                        s,
                                        website["url"],
                                        {
                                            "err_re_process": website["err_re_process"]
                                            + 1
                                        },
                                    ),
                                    read_concern=rc_snapshot,
                                    write_concern=wc_majority,
                                )

                            # with connection.start_session() as session:
                            #     with session.start_transaction():
                            #         collection.update_one(
                            #             {"url": website["url"]},
                            #             {
                            #                 "$set": {
                            #                     "err_re_process": website["err_re_process"] + 1
                            #                 },
                            #             },
                            #             session=session,
                            #         )

                        # ! check status if error and error > 2 time => kill process
                        if website["status"] == "error" and website[
                            "err_re_crawl"
                        ] > int(self.retry_time_limit):
                            self.get_logger().info(
                                f"[*] Remove cronjob of {website['spider_name']}: exceed retry time when crawl website"
                            )
                            cron.remove(job)
                            cron.write()
                            return

                        # check if modified schedule
                        if (
                            job.slices.render() != "* * * * *"
                            and job.slices.render() != website["schedule"]
                        ) or (
                            job.slices.render() == "* * * * *"
                            and website["schedule"].replace("/1", "")
                            != job.slices.render()
                        ):
                            current_schedule = job.slices.render()
                            self.get_logger().info(
                                f"[*] Change schedule time for {website['spider_name']}: from {current_schedule} to {website['schedule']}"
                            )
                            job.setall(website["schedule"])
                            cron.write()

                if not found:
                    # ! check status in-active if err_re_process > 2 time => omit
                    if website["status"] == "in-active":
                        if website["err_re_process"] > int(self.retry_time_limit):
                            return
                        # update err_re_process
                        with connection.start_session(
                            {"causalConsistency": False}
                        ) as session:
                            session.with_transaction(
                                lambda s: self.update_website_callback(
                                    s,
                                    website["url"],
                                    {"err_re_process": website["err_re_process"] + 1},
                                ),
                                read_concern=rc_snapshot,
                                write_concern=wc_majority,
                            )

                        # with connection.start_session() as session:
                        #     with session.start_transaction():
                        #         collection.update_one(
                        #             {"url": website["url"]},
                        #             {
                        #                 "$set": {
                        #                     "err_re_process": website["err_re_process"]
                        #                     + 1
                        #                 },
                        #             },
                        #             session=session,
                        #         )

                    # ! check status error if error > 2 time => omit
                    if website["status"] == "error" and website["err_re_crawl"] > int(
                        self.retry_time_limit
                    ):
                        return

                    # start new cronjob

                    # create new cron job based on template
                    job = cron.new(
                        command=self.cron_cmd_template.format(
                            website["spider_name"], website["url"]
                        ),
                        comment=website["spider_name"],
                    )

                    job.setall(website["schedule"])
                    cron.write()

                    self.get_logger().info(
                        f"[*] Init new job for {website['spider_name']}"
                    )

            connection.close()
        except Exception as error:
            if connection:
                connection.close()

            # traceback.print_exc()
            # + log to log server
            self.get_logger().error(
                f"[x] Error when process cronjob of website. \n"
                + traceback.format_exc()
            )
            sys.exit()

    def update_website_callback(self, session, url, update_obj):
        self.website_coll.update_one(
            {"url": url},
            {
                "$set": update_obj,
            },
            session=session,
        )



import scrapy

from .base import BaseItem


class WebsiteItem(BaseItem):
    name = scrapy.Field()
    css_selector = scrapy.Field()
    categories_num = scrapy.Field()
    category_urls = scrapy.Field()
    total_categories = scrapy.Field()



class BaseMessageBrokerConnector:
    def __init__(self, **kwargs):
        pass

    def send(self, message, target, **kwargs):
        pass

    def receive(self, receive_from, received_callback, **kwargs):
        pass

    def close(self):
        pass

    def send_processed_message_status(self, is_processed=True, **kwargs):
        pass


import os
import traceback

from pymongo import ReturnDocument, read_concern, write_concern

from ...pipelines.base import BasePipeline
from ...connectors.databases.mongodb_connector import MongoDBConnector
from ...connectors.message_brokers.rabbitmq_connector import RabbitMQConnector


class WebsitePipeline(BasePipeline):
    def __init__(self):
        self.receive_queue = None
        self.transmit_queue = None

        self.website_coll = None
        self.categories_coll = None

        super().__init__()

    def load_database_env(self):
        super().load_database_env()

    def setup_database(self, reinit=False):
        super().setup_database()
        self.website_coll = self.database[f'{os.getenv("WEBSITE_COLL")}']
        self.categories_coll = self.database[f'{os.getenv("CATEGORIES_COLL")}']

    def load_connector_env(self):
        super().load_connector_env()
        self.receive_queue = os.getenv("RECEIVE_QUEUE")
        self.connector_params["receive_queue"] = self.receive_queue
        self.transmit_queue = os.getenv("TRANSMIT_QUEUE")
        self.connector_params["transmit_queue"] = self.transmit_queue

    def init_database_connector(self):
        self.database_client = MongoDBConnector(
            connection_string=self.database_params["conn_string"],
            default_database=self.database_params["database"],
        )
        self.database = self.database_client.connection[
            self.database_params["database"]
        ]

    def init_message_connector(self):
        self.connector_client = RabbitMQConnector()

    def process_item(self, item, spider):
        # update status of website_data
        try:
            website = None

            rc_snapshot = read_concern.ReadConcern("snapshot")
            wc_majority = write_concern.WriteConcern("majority")
            # prepare update obj
            update_obj = None

            if item["status"] != "error":
                update_obj = {
                    "name": item["name"],
                    "brand": item["brand"],
                    "status": item["status"],
                    "last_crawl": item["last_crawl"],
                    "css_selector": item["css_selector"],
                    "categories_num": item["categories_num"],
                    "total_categories": item["total_categories"],
                    "err_re_crawl": 0,
                    "err_re_process": 0,  # clear err when cron init
                }
            else:
                website = self.website_coll.find_one({"url": item["url"]})

                update_obj = {
                    "name": item["name"],
                    "brand": item["brand"],
                    "status": item["status"],
                    "last_crawl": item["last_crawl"],
                    "total_categories": item["total_categories"],
                    "err_re_crawl": website["err_re_crawl"] + 1,
                    "err_re_process": 0,  # clear err when cron init
                }

            with self.database_client.connection.start_session(
                {"causalConsistency": False}
            ) as session:
                session.with_transaction(
                    lambda s: self.update_website_callback(s, item["url"], update_obj),
                    read_concern=rc_snapshot,
                    write_concern=wc_majority,
                )

            # with self.database_client.connection.start_session() as session:
            #     with session.start_transaction():
            #         update_obj = None

            #         if item["status"] != "error":
            #             update_obj = {
            #                 "name": item["name"],
            #                 "brand": item["brand"],
            #                 "status": item["status"],
            #                 "last_crawl": item["last_crawl"],
            #                 "css_selector": item["css_selector"],
            #                 "categories_num": item["categories_num"],
            #                 "total_categories": item["total_categories"],
            #                 "err_re_crawl": 0,
            #                 "err_re_process": 0,  # clear err when cron init
            #             }
            #         else:
            #             website = self.website_coll.find_one({"url": item["url"]})

            #             update_obj = {
            #                 "name": item["name"],
            #                 "brand": item["brand"],
            #                 "status": item["status"],
            #                 "last_crawl": item["last_crawl"],
            #                 "total_categories": item["total_categories"],
            #                 "err_re_crawl": website["err_re_crawl"] + 1,
            #                 "err_re_process": 0,  # clear err when cron init
            #             }

            #         website = self.website_coll.find_one_and_update(
            #             {"url": item["url"]},
            #             {
            #                 "$set": update_obj,
            #             },
            #             return_document=ReturnDocument.AFTER,
            #             session=session,
            #         )

            # send all categories_url to categories_queue
            website = self.website_coll.find_one({"url": item["url"]})

            if website:
                for category in item["category_urls"]:
                    # test rabbit
                    if category != item["category_urls"][0]:
                        continue

                    body_message = {
                        "website_id": str(website["_id"]),
                        "website_url": website["url"],
                        "website_name": website["name"],
                        "url": category,
                        "spider_name": website["spider_name"],
                        "is_root": "True",
                        "root_obj_id": "",
                    }

                    # send message to rabbitmq (categories_queue)
                    self.send_message(
                        body_message, self.connector_params["transmit_queue"]
                    )

            self.get_logger().info(
                f'[*] Crawling website {item["url"]} completed at {item["last_crawl"]}: status - "{item["status"]}"'
            )
        except Exception as error:
            # traceback.print_exc()
            # + log to log server
            self.get_logger().error(
                f'[x] Error when process data crawled in pipeline website (url: {item["url"]}). \n'
                + traceback.format_exc()
            )

        return item

    def close_spider(self, spider):
        # disconnect database
        self.database_client.close()

        # disconnect rabbitmq
        self.connector_client.close()

    def update_website_callback(self, session, url, update_obj):
        self.website_coll.update_one(
            {"url": url},
            {
                "$set": update_obj,
            },
            session=session,
        )



import json
import os
import sys
import random
import traceback
from datetime import datetime, timedelta
from pymongo import read_concern, write_concern

import asyncio
import psutil
from bson import ObjectId

from .base_crawler import BaseCrawler
from .connectors.databases.mongodb_connector import MongoDBConnector
from .connectors.message_brokers.rabbitmq_connector import RabbitMQConnector
from .log.default import get_default_log
from .mixins.spider_cache_mixin import SpiderCacheMixin


class GenericCrawler(BaseCrawler, SpiderCacheMixin):
    def __init__(self, name, coll_env_name):
        self.name = name
        self.coll_env_name = coll_env_name
        self.data_coll = None

        self.receive_queue = None
        self.transmit_queue = None

        self.website_coll = None
        self.categories_coll = None
        self.products_coll = None
        self.logger = None

        super().__init__()
        SpiderCacheMixin.__init__(self)

        self.user = os.getenv("USER")
        self.retry_time_limit = os.getenv("RETRY_TIME_LIMIT")

        # use phys_cpu
        # total_sub = psutil.cpu_count(logical=False) - 2
        self.total_sub_process = int(os.getenv("TOTAL_SUB_PROCESS"))

    def setup_database(self, reinit=False):
        super().setup_database(reinit=reinit)
        self.data_coll = self.database[f"{os.getenv(self.coll_env_name)}"]

    def init_database_connector(self):
        self.database_client = MongoDBConnector(
            connection_string=self.database_params["conn_string"],
            default_database=self.database_params["database"],
        )
        self.database = self.database_client.connection[
            self.database_params["database"]
        ]

    def load_connector_env(self):
        super().load_connector_env()
        self.receive_queue = os.getenv("RECEIVE_QUEUE")
        self.connector_params["receive_queue"] = self.receive_queue
        self.transmit_queue = os.getenv("TRANSMIT_QUEUE")
        self.connector_params["transmit_queue"] = self.transmit_queue

    def init_message_connector(self):
        self.connector_client = RabbitMQConnector(
            queue_declare=os.getenv("RECEIVE_QUEUE")
        )

    def get_logger(self):
        if not self.logger:
            self.logger = get_default_log(self.name)

        return self.logger

    def run(self):
        self.get_logger().info(f" [*] {self.name} instance: waiting for messages")
        self.receive_message(self.receive_queue, self.receive_message_callback)

    def receive_message_callback(self, ch, method, properties, body):
        self.get_logger().info(
            f"[*] Receive message in {self.name}_queue: (message: {body.decode()})"
        )

        rc_snapshot = read_concern.ReadConcern("snapshot")
        wc_majority = write_concern.WriteConcern("majority")

        body_message = None
        try:
            body_message = json.loads(body.decode())

            if not self.choose_spider(body_message["spider_name"]):
                return self.connector_client.send_processed_message_status(
                    is_processed=False, delivery_tag=method.delivery_tag
                )

            record = self.data_coll.find_one({"url": body_message["url"]})

            # ! check last crawl (not crawl document has last crawl > previous time range in env)
            # prevent continuos pagination categories
            if self.check_crawled_in_previous_time(record):
                return self.connector_client.send_processed_message_status(
                    is_processed=True, delivery_tag=method.delivery_tag
                )

            # ! check if status in-active and re process > retry time => omit (return ack)
            if self.check_process_retry_time(record):
                return self.connector_client.send_processed_message_status(
                    is_processed=True, delivery_tag=method.delivery_tag
                )

            #! check if status error and re crawl > retry time => omit (return ack)
            if self.check_error_retry_time(record):
                return self.connector_client.send_processed_message_status(
                    is_processed=True, delivery_tag=method.delivery_tag
                )

            if record:
                with self.database_client.connection.start_session(
                    {"causalConsistency": False}
                ) as session:
                    session.with_transaction(
                        lambda s: self.update_one_data_callback(
                            s,
                            {"url": body_message["url"]},
                            self.build_update_object(body_message),
                        ),
                        read_concern=rc_snapshot,
                        write_concern=wc_majority,
                    )

                # self.data_coll.update_one(
                #     {"url": body_message["url"]},
                #     {"$set": self.build_update_object(body_message)},
                #     session=session,
                # )
            else:
                # check process when receive new url of product instance
                if self.name == "product":
                    # 1. check unique product id
                    exist_product = self.data_coll.find_one(
                        {"unique_product_id": body_message["unique_product_id"]}
                    )

                    # 2 if has unique product id => get exist document => check 3 condition below => update url to url of exist document
                    if exist_product:
                        if self.check_crawled_in_previous_time(exist_product):
                            return self.connector_client.send_processed_message_status(
                                is_processed=True,
                                delivery_tag=method.delivery_tag,
                            )

                        if self.check_process_retry_time(exist_product):
                            return self.connector_client.send_processed_message_status(
                                is_processed=True,
                                delivery_tag=method.delivery_tag,
                            )

                        if self.check_error_retry_time(exist_product):
                            return self.connector_client.send_processed_message_status(
                                is_processed=True,
                                delivery_tag=method.delivery_tag,
                            )

                        with self.database_client.connection.start_session(
                            {"causalConsistency": False}
                        ) as session:
                            session.with_transaction(
                                lambda s: self.update_one_data_callback(
                                    s,
                                    {
                                        "unique_product_id": exist_product[
                                            "unique_product_id"
                                        ]
                                    },
                                    self.build_update_object(body_message),
                                ),
                                read_concern=rc_snapshot,
                                write_concern=wc_majority,
                            )

                        # self.data_coll.update_one(
                        #     {
                        #         "unique_product_id": exist_product[
                        #             "unique_product_id"
                        #         ]
                        #     },
                        #     {"$set": self.build_update_object(body_message)},
                        #     session=session,
                        # )
                    else:
                        # 3. if not has unique product id => create new document
                        with self.database_client.connection.start_session(
                            {"causalConsistency": False}
                        ) as session:
                            session.with_transaction(
                                lambda s: self.insert_one_data_callback(
                                    s, self.build_insert_object(body_message)
                                ),
                                read_concern=rc_snapshot,
                                write_concern=wc_majority,
                            )

                        # self.data_coll.insert_one(
                        #     self.build_insert_object(body_message),
                        #     session=session,
                        # )
                else:
                    with self.database_client.connection.start_session(
                        {"causalConsistency": False}
                    ) as session:
                        session.with_transaction(
                            lambda s: self.insert_one_data_callback(
                                s, self.build_insert_object(body_message)
                            ),
                            read_concern=rc_snapshot,
                            write_concern=wc_majority,
                        )

                    # self.data_coll.insert_one(
                    #     self.build_insert_object(body_message),
                    #     session=session,
                    # )

            # with self.database_client.connection.start_session() as session:
            #     with session.start_transaction():
            #         if record:
            #             self.data_coll.update_one(
            #                 {"url": body_message["url"]},
            #                 {"$set": self.build_update_object(body_message)},
            #                 session=session,
            #             )
            #         else:
            #             # check process when receive new url of product instance
            #             if self.name == "product":
            #                 # 1. check unique product id
            #                 exist_product = self.data_coll.find_one(
            #                     {"unique_product_id": body_message["unique_product_id"]}
            #                 )

            #                 # 2 if has unique product id => get exist document => check 3 condition below => update url to url of exist document
            #                 if exist_product:
            #                     if self.check_crawled_in_previous_time(exist_product):
            #                         return self.connector_client.send_processed_message_status(
            #                             is_processed=True,
            #                             delivery_tag=method.delivery_tag,
            #                         )

            #                     if self.check_process_retry_time(exist_product):
            #                         return self.connector_client.send_processed_message_status(
            #                             is_processed=True,
            #                             delivery_tag=method.delivery_tag,
            #                         )

            #                     if self.check_error_retry_time(exist_product):
            #                         return self.connector_client.send_processed_message_status(
            #                             is_processed=True,
            #                             delivery_tag=method.delivery_tag,
            #                         )

            #                     self.data_coll.update_one(
            #                         {
            #                             "unique_product_id": exist_product[
            #                                 "unique_product_id"
            #                             ]
            #                         },
            #                         {"$set": self.build_update_object(body_message)},
            #                         session=session,
            #                     )
            #                 else:
            #                     # 3. if not has unique product id => create new document
            #                     self.data_coll.insert_one(
            #                         self.build_insert_object(body_message),
            #                         session=session,
            #                     )
            #             else:
            #                 self.data_coll.insert_one(
            #                     self.build_insert_object(body_message),
            #                     session=session,
            #                 )

            # query must after transaction commit
            record = self.data_coll.find_one({"url": body_message["url"]})

            if not record:
                return self.connector_client.send_processed_message_status(
                    is_processed=False, delivery_tag=method.delivery_tag
                )

            # limit subprocess
            child_process = psutil.Process()
            # print("total subprocess", len(child_process.children()))
            if len(child_process.children()) >= self.total_sub_process:
                return self.connector_client.send_processed_message_status(
                    is_processed=False, delivery_tag=method.delivery_tag
                )

            with self.database_client.connection.start_session(
                {"causalConsistency": False}
            ) as session:
                session.with_transaction(
                    lambda s: self.update_one_data_callback(
                        s,
                        {"url": record["url"]},
                        {
                            "err_re_process": record["err_re_process"] + 1,
                        },
                    ),
                    read_concern=rc_snapshot,
                    write_concern=wc_majority,
                )

            # with self.database_client.connection.start_session() as session:
            #     with session.start_transaction():
            #         self.data_coll.update_one(
            #             {"url": record["url"]},
            #             {
            #                 "$set": {
            #                     "err_re_process": record["err_re_process"] + 1,
            #                 }
            #             },
            #             session=session,
            #         )

            asyncio.run(self.run_subprocess(record))

            # add count for spider
            self.increase_spider_count(record["spider_name"])

            self.connector_client.send_processed_message_status(
                is_processed=True, delivery_tag=method.delivery_tag
            )
        except Exception as error:
            if self.database_client.connection:
                self.database_client.close()

            # traceback.print_exc()
            # + log to log server
            self.get_logger().error(
                f"[x] Error when process message from {self.name}_queue (message: {body_message}). \n"
                + traceback.format_exc()
            )
            sys.exit()

    def check_crawled_in_previous_time(self, record):
        return (
            record
            and record["status"] == "active"
            and (
                record["last_crawl"]
                > datetime.now() - timedelta(hours=int(os.getenv("PREV_TIME")))
            )
        )

    def check_process_retry_time(self, record):
        return (
            record
            and record["status"] == "in-active"
            and record["err_re_process"] > int(os.getenv("RETRY_TIME_LIMIT"))
        )

    def check_error_retry_time(self, record):
        return (
            record
            and record["status"] == "error"
            and record["err_re_crawl"] > int(os.getenv("RETRY_TIME_LIMIT"))
        )

    async def run_subprocess(self, crawl_record):
        delay_time = crawl_record["random_sleep"]
        # print(f"Delay time {delay_time}")
        # await asyncio.sleep(delay_time)
        website_id = str(crawl_record["website_id"])

        command = self.get_spider_command(crawl_record)
        # print("subprocess", f"sleep {delay_time} && " + command.replace("&", "\&"))
        # process = Popen(
        #     command.replace("&", "\&"), shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE
        # )

        process = await asyncio.create_subprocess_shell(
            cmd=f"sleep {delay_time} && " + command.replace("&", "\&")
        )

        self.get_logger().info(
            f"[*] Sub process created: "
            + f"sleep {delay_time} && "
            + command.replace("&", "\&")
        )

    def get_spider_command(self, crawl_record):
        raise NotImplemented()

    def build_update_object(self, body_message):
        return {}

    def build_insert_object(self, body_message):
        update_obj = self.build_update_object(body_message)
        update_obj["err_re_crawl"] = 0
        update_obj["err_re_process"] = 0

        return update_obj

    def update_one_data_callback(self, session, query, update_obj):
        self.data_coll.update_one(
            query,
            {
                "$set": update_obj,
            },
            session=session,
        )

    def insert_one_data_callback(self, session, insert_obj):
        self.data_coll.insert_one(
            insert_obj,
            session=session,
        )
